/**
 * A first-in, first-out bounded collection of Message objects.
 */
public class BoundedQueue
{
    private Message elements[];
    private int head;
    private int tail;
    private int size;
    private boolean debug;

    /**
     * Constructs an empty queue.
     * @param capacity the maximum capacity of the queue
     */
    public BoundedQueue(int capacity) 
    {
        elements = new Message[capacity];
        head = 0;
        tail = 0;
        size = 0;
        debug = true;
    }
    
    public void setDebug(boolean aDebug) { debug = aDebug; }

    /**
     * Removes the object at the head.
     * @return the object that has been removed from the queue 
     * @precondition !isEmpty()
     */
    public synchronized Message remove() throws InterruptedException
    {
        while (isEmpty()) wait();
        Message msg = elements[head];

        head++;
        size--;
        if (head == elements.length) head = 0;

        if (debug) {
            msg.sequence = -msg.sequence;
            System.out.printf("%2d C %s\n", size, msg);
            msg.sequence = -msg.sequence;
        }

        notifyAll();
        return msg;
    }

    /**
     * Appends an object at the tail.
     * @param newValue the object to be appended 
     * @precondition !isFull();
     */
    public synchronized void add(Message newValue) throws InterruptedException
    {
        while (isFull()) wait();
        elements[tail] = newValue;

        tail++;
        size++;
        if (tail == elements.length) tail = 0;

        if (debug) System.out.printf("%2d P %s\n", size, newValue);
        notifyAll();
    }
    
    public boolean isFull() 
    {
        return size == elements.length;
    }

    public boolean isEmpty() 
    {
        return size == 0;
    }
}
